# Copyright (C) 2019  Nexedi SA and Contributors.
#                     Romain Courteaud <romain@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.

import unittest
from surykatka.db import LogDB
from playhouse.migrate import migrate, SqliteMigrator
from collections import namedtuple
from playhouse.reflection import Introspector

ValidationResult = namedtuple(
    "ValidationResult",
    (
        "valid",
        "table_exists",
        "add_fields",
        "remove_fields",
        "change_fields",
        "add_indexes",
        "remove_indexes",
    ),
)


def validate_schema(model):
    db = model._meta.database
    table = model._meta.table_name
    if not db.table_exists(table):
        return ValidationResult(False, False, None, None, None)

    introspector = Introspector.from_database(db)
    db_model = introspector.generate_models(table_names=[table])[table]

    columns = set(model._meta.columns)
    db_columns = set(db_model._meta.columns)

    to_remove = [model._meta.columns[c] for c in columns - db_columns]
    to_add = [db_model._meta.columns[c] for c in db_columns - columns]
    to_change = []
    intersect = columns & db_columns  # Take intersection and remove matches.
    for column in intersect:
        field = model._meta.columns[column]
        db_field = db_model._meta.columns[column]
        if (
            (field.field_type != db_field.field_type)
            and (not (field.field_type == "BIGINT"))
            and (not field.field_type == "JSON")
        ):
            to_change.append((field, db_field))

    indexes = set([(tuple(x), y) for x, y in model._meta.indexes])
    db_indexes = set(
        [
            (tuple(x), y)
            for x, y in db_model._meta.indexes
            if tuple(x) != model._meta.primary_key.field_names
        ]
    )
    index_to_remove = [c for c in indexes - db_indexes]
    index_to_add = [c for c in db_indexes - indexes]

    is_valid = not any(
        (to_remove, to_add, to_change, index_to_remove, index_to_add)
    )
    return ValidationResult(
        is_valid,
        True,
        to_add,
        to_remove,
        to_change,
        index_to_remove,
        index_to_add,
    )


class SurykatkaDBTestCase(unittest.TestCase):
    def setUp(self):
        self.db = LogDB(":memory:")

    def test_createTable(self):
        assert self.db._db.pragma("user_version") == 0
        self.db.createTables()
        assert self.db._db.pragma("user_version") == 5
        assert validate_schema(self.db.HttpCodeChange).valid, validate_schema(
            self.db.HttpCodeChange
        )
        assert validate_schema(self.db.SslChange).valid, validate_schema(
            self.db.SslChange
        )
        assert validate_schema(self.db.DnsChange).valid, validate_schema(
            self.db.DnsChange
        )
        assert validate_schema(self.db.NetworkChange).valid, validate_schema(
            self.db.NetworkChange
        )

    def test_downgrade(self):
        assert self.db._db.pragma("user_version") == 0
        # Unpossible high version
        self.db._db.pragma("user_version", 9999)
        try:
            self.db.createTables()
        except ValueError:
            assert self.db._db.pragma("user_version") == 9999
        else:
            raise NotImplementedError("Expected ValueError")

    def test_migrationFromVersion1(self):
        assert self.db._db.pragma("user_version") == 0
        # Recreate version 1
        with self.db._db.transaction():

            self.db._db.create_tables(
                [
                    self.db.Status,
                    self.db.ConfigurationChange,
                    self.db.HttpCodeChange,
                    self.db.NetworkChange,
                    self.db.PlatformChange,
                    self.db.DnsChange,
                ]
            )
            migrate(
                SqliteMigrator(self.db._db).drop_column(
                    "HttpCodeChange", "total_seconds"
                ),
            )
            migrate(
                SqliteMigrator(self.db._db).drop_column(
                    "HttpCodeChange", "http_header_dict"
                ),
            )
            migrate(
                SqliteMigrator(self.db._db).drop_index(
                    "HttpCodeChange", "HttpCodeChange_ip_url_status_id"
                ),
            )
            migrate(
                SqliteMigrator(self.db._db).drop_index(
                    "DnsChange",
                    "DnsChange_resolver_ip_domain_rdtype_status_id",
                ),
            )
            migrate(
                SqliteMigrator(self.db._db).drop_index(
                    "NetworkChange",
                    "NetworkChange_ip_transport_port_status_id",
                ),
            )
            self.db._db.pragma("user_version", 1)

        self.db.createTables()
        assert self.db._db.pragma("user_version") == 5
        assert validate_schema(self.db.HttpCodeChange).valid, validate_schema(
            self.db.HttpCodeChange
        )
        assert validate_schema(self.db.SslChange).valid, validate_schema(
            self.db.SslChange
        )
        assert validate_schema(self.db.DnsChange).valid, validate_schema(
            self.db.DnsChange
        )
        assert validate_schema(self.db.NetworkChange).valid, validate_schema(
            self.db.NetworkChange
        )

    def test_migrationFromVersion2(self):
        assert self.db._db.pragma("user_version") == 0
        # Recreate version 2
        with self.db._db.transaction():

            self.db._db.create_tables(
                [
                    self.db.Status,
                    self.db.ConfigurationChange,
                    self.db.HttpCodeChange,
                    self.db.NetworkChange,
                    self.db.PlatformChange,
                    self.db.DnsChange,
                    self.db.SslChange,
                ]
            )
            migrate(
                SqliteMigrator(self.db._db).drop_column(
                    "HttpCodeChange", "total_seconds"
                ),
            )
            migrate(
                SqliteMigrator(self.db._db).drop_column(
                    "HttpCodeChange", "http_header_dict"
                ),
            )
            migrate(
                SqliteMigrator(self.db._db).drop_index(
                    "HttpCodeChange", "HttpCodeChange_ip_url_status_id"
                ),
            )
            migrate(
                SqliteMigrator(self.db._db).drop_index(
                    "SslChange", "SslChange_ip_port_hostname_status_id"
                ),
            )
            migrate(
                SqliteMigrator(self.db._db).drop_index(
                    "DnsChange",
                    "DnsChange_resolver_ip_domain_rdtype_status_id",
                ),
            )
            migrate(
                SqliteMigrator(self.db._db).drop_index(
                    "NetworkChange",
                    "NetworkChange_ip_transport_port_status_id",
                ),
            )
            self.db._db.pragma("user_version", 2)

        self.db.createTables()
        assert self.db._db.pragma("user_version") == 5
        assert validate_schema(self.db.HttpCodeChange).valid, validate_schema(
            self.db.HttpCodeChange
        )
        assert validate_schema(self.db.SslChange).valid, validate_schema(
            self.db.SslChange
        )
        assert validate_schema(self.db.DnsChange).valid, validate_schema(
            self.db.DnsChange
        )
        assert validate_schema(self.db.NetworkChange).valid, validate_schema(
            self.db.NetworkChange
        )

    def test_migrationFromVersion3(self):
        assert self.db._db.pragma("user_version") == 0
        # Recreate version 3
        with self.db._db.transaction():

            self.db._db.create_tables(
                [
                    self.db.Status,
                    self.db.ConfigurationChange,
                    self.db.HttpCodeChange,
                    self.db.NetworkChange,
                    self.db.PlatformChange,
                    self.db.DnsChange,
                    self.db.SslChange,
                ]
            )
            migrate(
                SqliteMigrator(self.db._db).drop_column(
                    "HttpCodeChange", "http_header_dict"
                ),
            )
            migrate(
                SqliteMigrator(self.db._db).drop_index(
                    "HttpCodeChange", "HttpCodeChange_ip_url_status_id"
                ),
            )
            migrate(
                SqliteMigrator(self.db._db).drop_index(
                    "SslChange", "SslChange_ip_port_hostname_status_id"
                ),
            )
            migrate(
                SqliteMigrator(self.db._db).drop_index(
                    "DnsChange",
                    "DnsChange_resolver_ip_domain_rdtype_status_id",
                ),
            )
            migrate(
                SqliteMigrator(self.db._db).drop_index(
                    "NetworkChange",
                    "NetworkChange_ip_transport_port_status_id",
                ),
            )
            self.db._db.pragma("user_version", 3)

        self.db.createTables()
        assert self.db._db.pragma("user_version") == 5
        assert validate_schema(self.db.HttpCodeChange).valid, validate_schema(
            self.db.HttpCodeChange
        )
        assert validate_schema(self.db.SslChange).valid, validate_schema(
            self.db.SslChange
        )
        assert validate_schema(self.db.DnsChange).valid, validate_schema(
            self.db.DnsChange
        )
        assert validate_schema(self.db.NetworkChange).valid, validate_schema(
            self.db.NetworkChange
        )

    def test_migrationFromVersion4(self):
        assert self.db._db.pragma("user_version") == 0
        # Recreate version 3
        with self.db._db.transaction():

            self.db._db.create_tables(
                [
                    self.db.Status,
                    self.db.ConfigurationChange,
                    self.db.HttpCodeChange,
                    self.db.NetworkChange,
                    self.db.PlatformChange,
                    self.db.DnsChange,
                    self.db.SslChange,
                ]
            )
            migrate(
                SqliteMigrator(self.db._db).drop_index(
                    "HttpCodeChange", "HttpCodeChange_ip_url_status_id"
                ),
            )
            migrate(
                SqliteMigrator(self.db._db).drop_index(
                    "SslChange", "SslChange_ip_port_hostname_status_id"
                ),
            )
            migrate(
                SqliteMigrator(self.db._db).drop_index(
                    "DnsChange",
                    "DnsChange_resolver_ip_domain_rdtype_status_id",
                ),
            )
            migrate(
                SqliteMigrator(self.db._db).drop_index(
                    "NetworkChange",
                    "NetworkChange_ip_transport_port_status_id",
                ),
            )
            self.db._db.pragma("user_version", 4)

        self.db.createTables()
        assert self.db._db.pragma("user_version") == 5
        assert validate_schema(self.db.HttpCodeChange).valid, validate_schema(
            self.db.HttpCodeChange
        )
        assert validate_schema(self.db.SslChange).valid, validate_schema(
            self.db.SslChange
        )
        assert validate_schema(self.db.DnsChange).valid, validate_schema(
            self.db.DnsChange
        )
        assert validate_schema(self.db.NetworkChange).valid, validate_schema(
            self.db.NetworkChange
        )


def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(SurykatkaDBTestCase))
    return suite


if __name__ == "__main__":
    unittest.main(defaultTest="suite")
